/*
 * Copyright 2010 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

/*
 * @test
 * @bug 6964498
 * @summary Test localized bootstrap methods
 * @author jrose
 *
 * @library ..
 * @compile -source 7 -target 7 IndyLocalBootstrap.java
 */
//Not yet: @run main/othervm -XX:+EnableInvokeDynamic meth.IndyLocalBootstrap

/*
 * Standalone testing:
 * <code>
 * $ cd $MY_REPO_DIR/langtools
 * $ (cd make; make)
 * $ ./dist/bootstrap/bin/javac -d dist test/tools/javac/meth/IndyLocalBootstrap.java
 * $ ./dist/bootstrap/bin/javap -c -classpath dist meth.IndyLocalBootstrap
 * </code>
 */

package meth;

import java.dyn.*;

class IndyLocalBootstrap {
    public static void main(String... args) throws Throwable {
        new IndyLocalBootstrap().run(args);
    }
    void run(String... args) throws Throwable {
        System.out.println("starting");
        Foo.doFoo1();
        System.out.println("one test done");
        Foo.doFoo2();
        System.out.println("last one");
        doFoo3();
        System.out.println("all done");
        if (args.length > 0)
            BadCases.doBad1();
    }

    @BootstrapMethod(value=IndyLocalBootstrap.class, name="classBSM")
    static class Foo {
        static Object doFoo1() throws Throwable {
            return InvokeDynamic.foo2(2); //use classBSM
        }

        @BootstrapMethod(FancySite.class)
        static void doFoo2() throws Throwable {
            InvokeDynamic.<void>foo2("three"); //use FancySite
        }
    }

    // /* //BAD
    static //*/
    private Object classBSM(Object caller, Object name, Object type) {
        return new FancySite((Class)caller, (String)name, (MethodType)type, "IndyLocalBootstrap.classBSM");
    }

    private static class FancySite extends CallSite {
        // Save some random parameters:
        final String name;
        final String whence;
        private FancySite(Class<?> caller, String name, MethodType type) {
            this(caller, name, type, "FancySite.<init>");
        }
        private FancySite(Class<?> caller, String name, MethodType type, String whence) {
            this.name = name;
            this.whence = whence;
            MethodHandle handler = handler_MH.bindTo(this);  // = this#handler(...)
            setTarget(MethodHandles.collectArguments(handler, type));
        }

        private Object handler(Object... args) {
            System.out.println("Fancy call (BSM from "+whence+") "+name+getTarget().type()+" "+java.util.Arrays.asList(args));
            return null;
        }

        private static final MethodHandle handler_MH;
        static {
            MethodHandles.Lookup lookup = MethodHandles.lookup();
            handler_MH = // FancySite#handler(...)
                lookup.findVirtual(lookup.lookupClass(), "handler",
                    MethodType.methodType(Object.class, Object[].class));
        }

        // test inner class access to privates
        private static CallSite make(Class<?> caller, String name, MethodType type) {
            return new FancySite(caller, name, type, "FancySite.make");
        }
    }

    static Object doFoo3() throws Throwable {
        @BootstrapMethod(value=IndyLocalBootstrap.FancySite.class, name="make")
        Object junk = InvokeDynamic.foo3();
        return junk;
    }

    @BootstrapMethod(value=IndyLocalBootstrap.class, name="classBSM")
    static class BadCases {
        static Object doBad1() throws Throwable {
            try {
                @BootstrapMethod(value=java.util.Arrays.class, name="asList")
                Object junk = InvokeDynamic.bad1();
                throw new RuntimeException("unexpected return: "+junk);
            } catch (ClassCastException ex) {
                return ex;  // should reach here!
            }
        }

        //@BootstrapMethod(value=String.class, name="no.such.method") //BAD
        static void doBad2() throws Throwable {
            InvokeDynamic.bad2();
        }

        //@BootstrapMethod(value=String.class, name="toString") //BAD
        static void doBad3() throws Throwable {
            InvokeDynamic.bad3();
        }
    }
}
